/*!
* @file ILI9163.cpp
*
* @mainpage ILI9163 TFT Displays
*
* @section intro_sec Introduction
*
* This is the documentation for ILI9163 driver for the
* Arduino platform.
*/

#include "ILI9163.h"
#ifndef ARDUINO_STM32_FEATHER
  #include "pins_arduino.h"
  #ifndef RASPI
    #include "wiring_private.h"
  #endif
#endif
#include <limits.h>

#if defined (ARDUINO_ARCH_ARC32) || defined (ARDUINO_MAXIM)
  #define SPI_DEFAULT_FREQ  16000000
#elif defined (__AVR__) || defined(TEENSYDUINO)
  #define SPI_DEFAULT_FREQ  8000000
#elif defined(ESP8266) || defined(ESP32)
  #define SPI_DEFAULT_FREQ  40000000
#elif defined(RASPI)
  #define SPI_DEFAULT_FREQ  80000000
#elif defined(ARDUINO_ARCH_STM32F1)
  #define SPI_DEFAULT_FREQ  36000000
#else
  #define SPI_DEFAULT_FREQ  24000000  ///< Default SPI data clock frequency
#endif

#define MADCTL_MY  0x80  ///< Bottom to top
#define MADCTL_MX  0x40  ///< Right to left
#define MADCTL_MV  0x20  ///< Reverse Mode
#define MADCTL_ML  0x10  ///< LCD refresh Bottom to top
#define MADCTL_RGB 0x00  ///< Red-Green-Blue pixel order
#define MADCTL_BGR 0x08  ///< Blue-Green-Red pixel order
#define MADCTL_MH  0x04  ///< LCD refresh right to left

/**************************************************************************/
/*!
    @brief  Instantiate Adafruit ILI9163 driver with software SPI
    @param    cs    Chip select pin #
    @param    dc    Data/Command pin #
    @param    mosi  SPI MOSI pin #
    @param    sclk  SPI Clock pin #
    @param    rst   Reset pin # (optional, pass -1 if unused)
    @param    miso  SPI MISO pin # (optional, pass -1 if unused)
*/
/**************************************************************************/
ILI9163::ILI9163(int8_t cs, int8_t dc, int8_t mosi,
        int8_t sclk, int8_t rst, int8_t miso) : Adafruit_SPITFT(ILI9163_TFTWIDTH, ILI9163_TFTHEIGHT, cs, dc, mosi, sclk, rst, miso) {
}

/**************************************************************************/
/*!
    @brief  Instantiate Adafruit ILI9163 driver with hardware SPI using the
            default SPI peripheral.
    @param  cs   Chip select pin # (OK to pass -1 if CS tied to GND).
    @param  dc   Data/Command pin # (required).
    @param  rst  Reset pin # (optional, pass -1 if unused).
*/
/**************************************************************************/
ILI9163::ILI9163(int8_t cs, int8_t dc, int8_t rst) :
  Adafruit_SPITFT(ILI9163_TFTWIDTH, ILI9163_TFTHEIGHT, cs, dc, rst) {
}

#if !defined(ESP8266)
/**************************************************************************/
/*!
    @brief  Instantiate Adafruit ILI9163 driver with hardware SPI using
            a specific SPI peripheral (not necessarily default).
    @param  spiClass  Pointer to SPI peripheral (e.g. &SPI or &SPI1).
    @param  dc        Data/Command pin # (required).
    @param  cs        Chip select pin # (optional, pass -1 if unused and
                      CS is tied to GND).
    @param  rst       Reset pin # (optional, pass -1 if unused).
*/
/**************************************************************************/
ILI9163::ILI9163(
  SPIClass *spiClass, int8_t dc, int8_t cs, int8_t rst) :
  Adafruit_SPITFT(ILI9163_TFTWIDTH, ILI9163_TFTHEIGHT, spiClass, cs, dc, rst) {
}
#endif // end !ESP8266

//static const uint8_t PROGMEM initcmd[] = {
//  0xEF, 3, 0x03, 0x80, 0x02,
//  0xCF, 3, 0x00, 0xC1, 0x30,
//  0xED, 4, 0x64, 0x03, 0x12, 0x81,
//  0xE8, 3, 0x85, 0x00, 0x78,
//  0xCB, 5, 0x39, 0x2C, 0x00, 0x34, 0x02,
//  0xF7, 1, 0x20,
//  0xEA, 2, 0x00, 0x00,
//  ILI9163_PWCTR1  , 1, 0x23,             // Power control VRH[5:0]
//  ILI9163_PWCTR2  , 1, 0x10,             // Power control SAP[2:0];BT[3:0]
//  ILI9163_VMCTR1  , 2, 0x3e, 0x28,       // VCM control
//  ILI9163_VMCTR2  , 1, 0x86,             // VCM control2
//  ILI9163_MADCTL  , 1, 0x48,             // Memory Access Control
//  ILI9163_VSCRSADD, 1, 0x00,             // Vertical scroll zero
//  ILI9163_PIXFMT  , 1, 0x55,
//  ILI9163_FRMCTR1 , 2, 0x00, 0x18,
//  ILI9163_DFUNCTR , 3, 0x08, 0x82, 0x27, // Display Function Control
//  0xF2, 1, 0x00,                         // 3Gamma Function Disable
//  ILI9163_GAMMASET , 1, 0x01,             // Gamma curve selected
//  ILI9163_GMCTRP1 , 15, 0x0F, 0x31, 0x2B, 0x0C, 0x0E, 0x08, // Set Gamma
//    0x4E, 0xF1, 0x37, 0x07, 0x10, 0x03, 0x0E, 0x09, 0x00,
//  ILI9163_GMCTRN1 , 15, 0x00, 0x0E, 0x14, 0x03, 0x11, 0x07, // Set Gamma
//    0x31, 0xC1, 0x48, 0x08, 0x0F, 0x0C, 0x31, 0x36, 0x0F,
//  ILI9163_SLPOUT  , 0x80,                // Exit Sleep
//  ILI9163_DISPON  , 0x80,                // Display on
//  0x00                                   // End of list
//};

static const uint8_t PROGMEM initcmd[] = {
  0x11, 0x80,
  0x26, 1, 0x04,
  0xf2, 1, 0x00,
  0xb1, 2, 0x0a, 0x14,
//  0xb4, 1, 0x07, // Frame inversion
  0xb4, 1, 0x00, // Line inversion
  0xc0, 2, 0x0a, 0x00,
  0xc1, 1, 0x02,
  0xc5, 2, 0x2f, 0x3e,
  0xc7, 1, 0x40,
  0x2a, 4, 0x00, 0x00, 0x00, 0x7f,
  0x2b, 4, 0x00, 0x00, 0x00, 0x9f,
  0x36, 1, 0x00,
  0x37, 2, 0x00, 0x00,
//  0x3a, 1, 0x66,
  0x3a, 1, 0x55,
  0x29, 0x80,
  0x2c, 0x80,
  0x00                                   // End of list
};

/**************************************************************************/
/*!
    @brief   Initialize ILI9163 chip
    Connects to the ILI9163 over SPI and sends initialization procedure commands
    @param    freq  Desired SPI clock frequency
*/
/**************************************************************************/
void ILI9163::begin(uint32_t freq) {

    if(!freq) freq = SPI_DEFAULT_FREQ;
    initSPI(freq);

    startWrite();

    if(_rst < 0) {                     // If no hardware reset pin...
        writeCommand(ILI9163_SWRESET); // Engage software reset
        delay(150);
    }

    uint8_t        cmd, x, numArgs;
    const uint8_t *addr = initcmd;
    while((cmd = pgm_read_byte(addr++)) > 0) {
        writeCommand(cmd);
        x       = pgm_read_byte(addr++);
        numArgs = x & 0x7F;
        while(numArgs--) spiWrite(pgm_read_byte(addr++));
        if(x & 0x80) delay(150);
    }

    endWrite();

    _width  = ILI9163_TFTWIDTH;
    _height = ILI9163_TFTHEIGHT;
}


/**************************************************************************/
/*!
    @brief   Set origin of (0,0) and orientation of TFT display
    @param   m  The index for rotation, from 0-3 inclusive
*/
/**************************************************************************/
void ILI9163::setRotation(uint8_t m) {
    rotation = m % 8;
    switch (rotation) {
        case 0:
//            m = (MADCTL_MX | MADCTL_BGR);
            m = (MADCTL_RGB);
            _width  = ILI9163_TFTWIDTH;
            _height = ILI9163_TFTHEIGHT;
            break;
        case 1:
//            m = (MADCTL_MV | MADCTL_BGR);
            m = (MADCTL_MX | MADCTL_MV | MADCTL_RGB);
            _width  = ILI9163_TFTHEIGHT;
            _height = ILI9163_TFTWIDTH;
            break;
        case 2:
//            m = (MADCTL_MY | MADCTL_BGR);
            m = (MADCTL_MX | MADCTL_MY | MADCTL_RGB);
            _width  = ILI9163_TFTWIDTH;
            _height = ILI9163_TFTHEIGHT;
            break;
        case 3:
//            m = (MADCTL_MX | MADCTL_MY | MADCTL_MV | MADCTL_BGR);
            m = (MADCTL_MY | MADCTL_MV | MADCTL_RGB);
            _width  = ILI9163_TFTHEIGHT;
            _height = ILI9163_TFTWIDTH;
            break;
        case 4:
            m = (MADCTL_BGR);
            _width  = ILI9163_TFTWIDTH;
            _height = ILI9163_TFTHEIGHT;
            break;
        case 5:
            m = (MADCTL_MX | MADCTL_MV | MADCTL_BGR);
            _width  = ILI9163_TFTHEIGHT;
            _height = ILI9163_TFTWIDTH;
            break;
        case 6:
            m = (MADCTL_MX | MADCTL_MY | MADCTL_BGR);
            _width  = ILI9163_TFTWIDTH;
            _height = ILI9163_TFTHEIGHT;
            break;
        case 7:
            m = (MADCTL_MY | MADCTL_MV | MADCTL_BGR);
            _width  = ILI9163_TFTHEIGHT;
            _height = ILI9163_TFTWIDTH;
            break;
    }

    startWrite();
    writeCommand(ILI9163_MADCTL);
    spiWrite(m);
    endWrite();
}

/**************************************************************************/
/*!
    @brief   Enable/Disable display color inversion
    @param   invert True to invert, False to have normal color
*/
/**************************************************************************/
void ILI9163::invertDisplay(bool invert) {
    startWrite();
    writeCommand(invert ? ILI9163_INVON : ILI9163_INVOFF);
    endWrite();
}

/**************************************************************************/
/*!
    @brief   Scroll display memory
    @param   y How many pixels to scroll display by
*/
/**************************************************************************/
void ILI9163::scrollTo(uint16_t y) {
    startWrite();
    writeCommand(ILI9163_VSCRSADD);
    SPI_WRITE16(y);
    endWrite();
}

/**************************************************************************/
/*!
    @brief   Set the "address window" - the rectangle we will write to RAM with the next chunk of SPI data writes. The ILI9163 will automatically wrap the data as each row is filled
    @param   x1  TFT memory 'x' origin
    @param   y1  TFT memory 'y' origin
    @param   w   Width of rectangle
    @param   h   Height of rectangle
*/
/**************************************************************************/
void ILI9163::setAddrWindow(
  uint16_t x1, uint16_t y1, uint16_t w, uint16_t h) {
    uint16_t x2 = (x1 + w - 1),
             y2 = (y1 + h - 1);
    writeCommand(ILI9163_CASET); // Column address set
    SPI_WRITE16(x1);
    SPI_WRITE16(x2);
    writeCommand(ILI9163_PASET); // Row address set
    SPI_WRITE16(y1);
    SPI_WRITE16(y2);
    writeCommand(ILI9163_RAMWR); // Write to RAM
}

/**************************************************************************/
/*!
   @brief  Read 8 bits of data from ILI9163 configuration memory. NOT from RAM!
           This is highly undocumented/supported, it's really a hack but kinda works?
    @param    command  The command register to read data from
    @param    index  The byte index into the command to read from
    @return   Unsigned 8-bit data read from ILI9163 register
*/
/**************************************************************************/
uint8_t ILI9163::readcommand8(uint8_t command, uint8_t index) {
    startWrite();
    writeCommand(0xD9);  // woo sekret command?
    spiWrite(0x10 + index);
    writeCommand(command);
    uint8_t r = spiRead();
    endWrite();
    return r;
}

