/*!
* ILI9163_buf.cpp
*   ILI9163 TFT Displays with frame buffer
*
* Caution. This library use 40,960byte RAM for frame buffer
*
* Yuji Tanaka, Japan
*
*/

#include "ILI9163_buf.h"
#ifndef ARDUINO_STM32_FEATHER
  #include "pins_arduino.h"
  #ifndef RASPI
    #include "wiring_private.h"
  #endif
#endif
#include <limits.h>

#if defined (ARDUINO_ARCH_ARC32) || defined (ARDUINO_MAXIM)
  #define SPI_DEFAULT_FREQ  16000000
#elif defined (__AVR__) || defined(TEENSYDUINO)
  #define SPI_DEFAULT_FREQ  8000000
#elif defined(ESP8266) || defined(ESP32)
  #define SPI_DEFAULT_FREQ  40000000
#elif defined(RASPI)
  #define SPI_DEFAULT_FREQ  80000000
#elif defined(ARDUINO_ARCH_STM32F1)
  #define SPI_DEFAULT_FREQ  36000000
#else
  #define SPI_DEFAULT_FREQ  24000000  ///< Default SPI data clock frequency
#endif

#define MADCTL_MY  0x80  ///< Bottom to top
#define MADCTL_MX  0x40  ///< Right to left
#define MADCTL_MV  0x20  ///< Reverse Mode
#define MADCTL_ML  0x10  ///< LCD refresh Bottom to top
#define MADCTL_RGB 0x00  ///< Red-Green-Blue pixel order
#define MADCTL_BGR 0x08  ///< Blue-Green-Red pixel order
#define MADCTL_MH  0x04  ///< LCD refresh right to left

#ifndef _swap_int16_t
#define _swap_int16_t(a, b) { int16_t t = a; a = b; b = t; }
#endif

/**************************************************************************/
/*!
    @brief  Instantiate Adafruit ILI9163 driver with software SPI
    @param    cs    Chip select pin #
    @param    dc    Data/Command pin #
    @param    mosi  SPI MOSI pin #
    @param    sclk  SPI Clock pin #
    @param    rst   Reset pin # (optional, pass -1 if unused)
    @param    miso  SPI MISO pin # (optional, pass -1 if unused)
*/
/**************************************************************************/
ILI9163_buf::ILI9163_buf(int8_t cs, int8_t dc, int8_t mosi,
        int8_t sclk, int8_t rst, int8_t miso) : Adafruit_SPITFT(ILI9163_TFTWIDTH, ILI9163_TFTHEIGHT, cs, dc, mosi, sclk, rst, miso) {
}

/**************************************************************************/
/*!
    @brief  Instantiate Adafruit ILI9163 driver with hardware SPI using the
            default SPI peripheral.
    @param  cs   Chip select pin # (OK to pass -1 if CS tied to GND).
    @param  dc   Data/Command pin # (required).
    @param  rst  Reset pin # (optional, pass -1 if unused).
*/
/**************************************************************************/
ILI9163_buf::ILI9163_buf(int8_t cs, int8_t dc, int8_t rst) :
  Adafruit_SPITFT(ILI9163_TFTWIDTH, ILI9163_TFTHEIGHT, cs, dc, rst) {
}

#if !defined(ESP8266)
/**************************************************************************/
/*!
    @brief  Instantiate Adafruit ILI9163 driver with hardware SPI using
            a specific SPI peripheral (not necessarily default).
    @param  spiClass  Pointer to SPI peripheral (e.g. &SPI or &SPI1).
    @param  dc        Data/Command pin # (required).
    @param  cs        Chip select pin # (optional, pass -1 if unused and
                      CS is tied to GND).
    @param  rst       Reset pin # (optional, pass -1 if unused).
*/
/**************************************************************************/
ILI9163_buf::ILI9163_buf(
  SPIClass *spiClass, int8_t dc, int8_t cs, int8_t rst) :
  Adafruit_SPITFT(ILI9163_TFTWIDTH, ILI9163_TFTHEIGHT, spiClass, cs, dc, rst) {
}
#endif // end !ESP8266

static const uint8_t PROGMEM initcmd[] = {
  0x11, 0x80,
  0x26, 1, 0x04,
  0xf2, 1, 0x00,
  0xb1, 2, 0x0a, 0x14,
//  0xb4, 1, 0x07, // Frame inversion
  0xb4, 1, 0x00, // Line inversion
  0xc0, 2, 0x0a, 0x00,
  0xc1, 1, 0x02,
  0xc5, 2, 0x2f, 0x3e,
  0xc7, 1, 0x40,
  0x2a, 4, 0x00, 0x00, 0x00, 0x7f,
  0x2b, 4, 0x00, 0x00, 0x00, 0x9f,
  0x36, 1, 0x00,
  0x37, 2, 0x00, 0x00,
//  0x3a, 1, 0x66,
  0x3a, 1, 0x55,
  0x29, 0x80,
  0x2c, 0x80,
  0x00                                   // End of list
};

/**************************************************************************/
/*!
    @brief   Initialize ILI9163 chip
    Connects to the ILI9163 over SPI and sends initialization procedure commands
    @param    freq  Desired SPI clock frequency
*/
/**************************************************************************/
void ILI9163_buf::begin(uint32_t freq) {

    if(!freq) freq = SPI_DEFAULT_FREQ;
    initSPI(freq);

    startWrite();

    if(_rst < 0) {                     // If no hardware reset pin...
        writeCommand(ILI9163_SWRESET); // Engage software reset
        delay(150);
    }

    uint8_t        cmd, x, numArgs;
    const uint8_t *addr = initcmd;
    while((cmd = pgm_read_byte(addr++)) > 0) {
        writeCommand(cmd);
        x       = pgm_read_byte(addr++);
        numArgs = x & 0x7F;
        while(numArgs--) spiWrite(pgm_read_byte(addr++));
        if(x & 0x80) delay(150);
    }

    endWrite();

    _width  = ILI9163_TFTWIDTH;
    _height = ILI9163_TFTHEIGHT;

    buffer = (uint8_t *)malloc(_width * _height * 2);
}


/**************************************************************************/
/*!
    @brief   Set origin of (0,0) and orientation of TFT display
    @param   m  The index for rotation, from 0-3 inclusive
*/
/**************************************************************************/
void ILI9163_buf::setRotation(uint8_t m) {
    rotation = m % 8;
    switch (rotation) {
        case 0:
//            m = (MADCTL_MX | MADCTL_BGR);
            m = (MADCTL_RGB);
            _width  = ILI9163_TFTWIDTH;
            _height = ILI9163_TFTHEIGHT;
            break;
        case 1:
//            m = (MADCTL_MV | MADCTL_BGR);
            m = (MADCTL_MX | MADCTL_MV | MADCTL_RGB);
            _width  = ILI9163_TFTHEIGHT;
            _height = ILI9163_TFTWIDTH;
            break;
        case 2:
//            m = (MADCTL_MY | MADCTL_BGR);
            m = (MADCTL_MX | MADCTL_MY | MADCTL_RGB);
            _width  = ILI9163_TFTWIDTH;
            _height = ILI9163_TFTHEIGHT;
            break;
        case 3:
//            m = (MADCTL_MX | MADCTL_MY | MADCTL_MV | MADCTL_BGR);
            m = (MADCTL_MY | MADCTL_MV | MADCTL_RGB);
            _width  = ILI9163_TFTHEIGHT;
            _height = ILI9163_TFTWIDTH;
            break;
        case 4:
            m = (MADCTL_BGR);
            _width  = ILI9163_TFTWIDTH;
            _height = ILI9163_TFTHEIGHT;
            break;
        case 5:
            m = (MADCTL_MX | MADCTL_MV | MADCTL_BGR);
            _width  = ILI9163_TFTHEIGHT;
            _height = ILI9163_TFTWIDTH;
            break;
        case 6:
            m = (MADCTL_MX | MADCTL_MY | MADCTL_BGR);
            _width  = ILI9163_TFTWIDTH;
            _height = ILI9163_TFTHEIGHT;
            break;
        case 7:
            m = (MADCTL_MY | MADCTL_MV | MADCTL_BGR);
            _width  = ILI9163_TFTHEIGHT;
            _height = ILI9163_TFTWIDTH;
            break;
    }

    startWrite();
    writeCommand(ILI9163_MADCTL);
    spiWrite(m);
    endWrite();
}

/**************************************************************************/
/*!
    @brief   Enable/Disable display color inversion
    @param   invert True to invert, False to have normal color
*/
/**************************************************************************/
void ILI9163_buf::invertDisplay(bool invert) {
    startWrite();
    writeCommand(invert ? ILI9163_INVON : ILI9163_INVOFF);
    endWrite();
}

/**************************************************************************/
/*!
    @brief   Scroll display memory
    @param   y How many pixels to scroll display by
*/
/**************************************************************************/
void ILI9163_buf::scrollTo(uint16_t y) {
    startWrite();
    writeCommand(ILI9163_VSCRSADD);
    SPI_WRITE16(y);
    endWrite();
}

/**************************************************************************/
/*!
    @brief   Set the "address window" - the rectangle we will write to RAM with the next chunk of SPI data writes. The ILI9163 will automatically wrap the data as each row is filled
    @param   x1  TFT memory 'x' origin
    @param   y1  TFT memory 'y' origin
    @param   w   Width of rectangle
    @param   h   Height of rectangle
*/
/**************************************************************************/
void ILI9163_buf::setAddrWindow(
  uint16_t x1, uint16_t y1, uint16_t w, uint16_t h) {
    uint16_t x2 = (x1 + w - 1),
             y2 = (y1 + h - 1);
    writeCommand(ILI9163_CASET); // Column address set
    SPI_WRITE16(x1);
    SPI_WRITE16(x2);
    writeCommand(ILI9163_PASET); // Row address set
    SPI_WRITE16(y1);
    SPI_WRITE16(y2);
    writeCommand(ILI9163_RAMWR); // Write to RAM
}

/**************************************************************************/
/*!
   @brief  Read 8 bits of data from ILI9163 configuration memory. NOT from RAM!
           This is highly undocumented/supported, it's really a hack but kinda works?
    @param    command  The command register to read data from
    @param    index  The byte index into the command to read from
    @return   Unsigned 8-bit data read from ILI9163 register
*/
/**************************************************************************/
uint8_t ILI9163_buf::readcommand8(uint8_t command, uint8_t index) {
    startWrite();
    writeCommand(0xD9);  // woo sekret command?
    spiWrite(0x10 + index);
    writeCommand(command);
    uint8_t r = spiRead();
    endWrite();
    return r;
}
/*
void ILI9163_buf::display() {
    uint16_t i;
    uint16_t val;
    uint8_t *buf;

    startWrite();
    setAddrWindow(0, 0, _width, _height);
    buf=buffer;
    for(i=0; i<_width * _height; i++){
        val=*(buf++) + *(buf++) * 256;
        SPI_WRITE16(val);
    }
    endWrite();
}
*/
/*
void ILI9163_buf::display() {
    uint16_t i;
    uint32_t val;
    uint8_t *buf;

    startWrite();
    setAddrWindow(0, 0, _width, _height);
    buf=buffer;
    for(i=0; i<_width * _height / 2; i++){
        val=(*(buf++) + *(buf++) * 256) * 65536 + *(buf++) + *(buf++) * 256;
        SPI_WRITE32(val);
    }
    endWrite();
}
*/
void ILI9163_buf::display() {
    startWrite();
    setAddrWindow(0, 0, _width, _height);
    writePixels((uint16_t *)buffer, _width * _height);
    endWrite();
}

void ILI9163_buf::clearDisplay() {
    memset(buffer, 0, _width * _height * 2);
}

void ILI9163_buf::writePixel(int16_t x, int16_t y, uint16_t color) {
    uint16_t i;

    if (x<0 || x>=_width || y<0 || y>=_height)
        return;

    i=((y * _width) + x) * 2;
    buffer[i]=color % 256;
    buffer[i+1]=color / 256;
}

void ILI9163_buf::writeFillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color) {
    uint16_t i;
    int16_t  x0, y0;
    int16_t  x1, y1;
    int16_t  px, py;

    if ((x+w)<0 || x>=_width || (y+h)<0 || y>=_height)
        return;

    x0=x;
    x1=x+w;
    y0=y;
    y1=y+h;

    if (x<0)
      x0=0;
    if ((x+w)>_width)
      x1=_width;

    if (y<0)
      y0=0;
    if ((y+h)>_height)
      y1=_height;

    for(py=y0; py<y1; py++){
      for(px=x0; px<x1; px++){
        i=((py * _width) + px) * 2;
        buffer[i]=color % 256;
        buffer[i+1]=color / 256;
      }
    }
}

void ILI9163_buf::writeFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color) {
    uint16_t i;
    int16_t  y0, y1;
    int16_t  px, py;

    if (x<0 || x>=_width || (y+h)<0 || y>=_height)
        return;

    y0=y;
    y1=y+h;

    if (y<0)
      y0=0;
    if ((y+h)>_height)
      y1=_height;

    px=x;
    for(py=y0; py<y1; py++){
        i=((py * _width) + px) * 2;
        buffer[i]=color % 256;
        buffer[i+1]=color / 256;
    }
}

void ILI9163_buf::writeFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color) {
    uint16_t i;
    int16_t  x0, x1;
    int16_t  px, py;

    if ((x+w)<0 || x>=_width || y<0 || y>=_height)
        return;

    x0=x;
    x1=x+w;

    if (x<0)
      x0=0;
    if ((x+w)>_width)
      x1=_width;

    py=y;
    for(px=x0; px<x1; px++){
        i=((py * _width) + px) * 2;
        buffer[i]=color % 256;
        buffer[i+1]=color / 256;
    }
}

void ILI9163_buf::drawPixel(int16_t x, int16_t y, uint16_t color) {
    writePixel(x, y, color);
}

void ILI9163_buf::drawRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color) {
    writeFastVLine(x, y, h, color);
    writeFastVLine(x + w, y, h, color);
    writeFastHLine(x, y, w, color);
    writeFastHLine(x, y + h, w, color);
}

void ILI9163_buf::fillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color) {
    writeFillRect(x, y, w, h, color);
}

void ILI9163_buf::drawFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color) {
    writeFastVLine(x, y, h, color);
}

void ILI9163_buf::drawFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color) {
    writeFastHLine(x, y, w, color);
}

void ILI9163_buf::drawRGBBitmap(int16_t x, int16_t y, const uint16_t bitmap[], int16_t w, int16_t h, int16_t dir) {
	int val;
	if (dir==0){
		for(int16_t j=0; j<h; j++, y++) {
			for(int16_t i=0; i<w; i++ ) {
				val=bitmap[j * w + i];
				if (val!=ILI9163_THROUGH)
					writePixel(x+i, y, val);
			}
		}
	}
	if (dir==4){
		for(int16_t j=0; j<h; j++, y++) {
			for(int16_t i=0; i<w; i++ ) {
				val=bitmap[j * w + i];
				if (val!=ILI9163_THROUGH)
					writePixel(x+w-1-i, y, val);
			}
		}
	}
	if (dir==8){
		for(int16_t j=0; j<h; j++) {
			for(int16_t i=0; i<w; i++ ) {
				val=bitmap[j * w + i];
				if (val!=ILI9163_THROUGH)
					writePixel(x+i, y+h-1-j, val);
			}
		}
	}
}
