/*!
* ILI9163_buf.cpp
*   ILI9163 TFT Displays with frame buffer
*
* Caution. This library use 40,960byte RAM for frame buffer
*
* Yuji Tanaka, Japan
*
*/

#ifndef _ILI9163_buf_H_
#define _ILI9163_buf_H_

#include "Arduino.h"
#include "Print.h"
#include <SPI.h>
#include "Adafruit_GFX.h"
#include <Adafruit_SPITFT.h>
#include <Adafruit_SPITFT_Macros.h>

#define ILI9163_TFTWIDTH   128      ///< ILI9163 max TFT width
#define ILI9163_TFTHEIGHT  160      ///< ILI9163 max TFT height

#define ILI9163_NOP        0x00     ///< No-op register
#define ILI9163_SWRESET    0x01     ///< Software reset register
#define ILI9163_RDDID      0x04     ///< Read display identification information
#define ILI9163_RDDST      0x09     ///< Read Display Status

#define ILI9163_SLPIN      0x10     ///< Enter Sleep Mode
#define ILI9163_SLPOUT     0x11     ///< Sleep Out
#define ILI9163_PTLON      0x12     ///< Partial Mode ON
#define ILI9163_NORON      0x13     ///< Normal Display Mode ON

#define ILI9163_RDMODE     0x0A     ///< Read Display Power Mode
#define ILI9163_RDMADCTL   0x0B     ///< Read Display MADCTL
#define ILI9163_RDPIXFMT   0x0C     ///< Read Display Pixel Format
#define ILI9163_RDIMGFMT   0x0D     ///< Read Display Image Format
#define ILI9163_RDSELFDIAG 0x0F     ///< Read Display Self-Diagnostic Result

#define ILI9163_INVOFF     0x20     ///< Display Inversion OFF
#define ILI9163_INVON      0x21     ///< Display Inversion ON
#define ILI9163_GAMMASET   0x26     ///< Gamma Set
#define ILI9163_DISPOFF    0x28     ///< Display OFF
#define ILI9163_DISPON     0x29     ///< Display ON

#define ILI9163_CASET      0x2A     ///< Column Address Set
#define ILI9163_PASET      0x2B     ///< Page Address Set
#define ILI9163_RAMWR      0x2C     ///< Memory Write
#define ILI9163_RAMRD      0x2E     ///< Memory Read

#define ILI9163_PTLAR      0x30     ///< Partial Area
#define ILI9163_MADCTL     0x36     ///< Memory Access Control
#define ILI9163_VSCRSADD   0x37     ///< Vertical Scrolling Start Address
#define ILI9163_PIXFMT     0x3A     ///< COLMOD: Pixel Format Set

#define ILI9163_FRMCTR1    0xB1     ///< Frame Rate Control (In Normal Mode/Full Colors)
#define ILI9163_FRMCTR2    0xB2     ///< Frame Rate Control (In Idle Mode/8 colors)
#define ILI9163_FRMCTR3    0xB3     ///< Frame Rate control (In Partial Mode/Full Colors)
#define ILI9163_INVCTR     0xB4     ///< Display Inversion Control
#define ILI9163_DFUNCTR    0xB6     ///< Display Function Control

#define ILI9163_PWCTR1     0xC0     ///< Power Control 1
#define ILI9163_PWCTR2     0xC1     ///< Power Control 2
#define ILI9163_PWCTR3     0xC2     ///< Power Control 3
#define ILI9163_PWCTR4     0xC3     ///< Power Control 4
#define ILI9163_PWCTR5     0xC4     ///< Power Control 5
#define ILI9163_VMCTR1     0xC5     ///< VCOM Control 1
#define ILI9163_VMCTR2     0xC7     ///< VCOM Control 2

#define ILI9163_RDID1      0xDA     ///< Read ID 1
#define ILI9163_RDID2      0xDB     ///< Read ID 2
#define ILI9163_RDID3      0xDC     ///< Read ID 3
#define ILI9163_RDID4      0xDD     ///< Read ID 4

#define ILI9163_GMCTRP1    0xE0     ///< Positive Gamma Correction
#define ILI9163_GMCTRN1    0xE1     ///< Negative Gamma Correction
//#define ILI9163_PWCTR6     0xFC

// Color definitions
#define ILI9163_BLACK       0x0000  ///<   0,   0,   0
#define ILI9163_NAVY        0x000F  ///<   0,   0, 123
#define ILI9163_DARKGREEN   0x03E0  ///<   0, 125,   0
#define ILI9163_DARKCYAN    0x03EF  ///<   0, 125, 123
#define ILI9163_MAROON      0x7800  ///< 123,   0,   0
#define ILI9163_PURPLE      0x780F  ///< 123,   0, 123
#define ILI9163_OLIVE       0x7BE0  ///< 123, 125,   0
#define ILI9163_LIGHTGREY   0xC618  ///< 198, 195, 198
#define ILI9163_DARKGREY    0x7BEF  ///< 123, 125, 123
#define ILI9163_BLUE        0x001F  ///<   0,   0, 255
#define ILI9163_GREEN       0x07E0  ///<   0, 255,   0
#define ILI9163_CYAN        0x07FF  ///<   0, 255, 255
#define ILI9163_RED         0xF800  ///< 255,   0,   0
#define ILI9163_MAGENTA     0xF81F  ///< 255,   0, 255
#define ILI9163_YELLOW      0xFFE0  ///< 255, 255,   0
#define ILI9163_WHITE       0xFFFF  ///< 255, 255, 255
#define ILI9163_ORANGE      0xFD20  ///< 255, 165,   0
#define ILI9163_GREENYELLOW 0xAFE5  ///< 173, 255,  41
#define ILI9163_PINK        0xFC18  ///< 255, 130, 198

#define ILI9163_THROUGH     0x0001  ///<   0,   0,   1

/**************************************************************************/
/*!
@brief Class to manage hardware interface with ILI9163 chipset (also seems to work with ILI9340)
*/
/**************************************************************************/

class ILI9163_buf : public Adafruit_SPITFT {
    public:
        ILI9163_buf(int8_t _CS, int8_t _DC, int8_t _MOSI, int8_t _SCLK,
          int8_t _RST = -1, int8_t _MISO = -1);
        ILI9163_buf(int8_t _CS, int8_t _DC, int8_t _RST = -1);
#if !defined(ESP8266)
        ILI9163_buf(SPIClass *spiClass, int8_t dc,
          int8_t cs = -1, int8_t rst = -1);
#endif // end !ESP8266

        void    begin(uint32_t freq=0);
        void    display(void);
        void    clearDisplay(void);
        void    writePixel(int16_t x, int16_t y, uint16_t color);
        void    writeFillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color);
        void    writeFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color);
        void    writeFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color);
        void    drawPixel(int16_t x, int16_t y, uint16_t color);
        void    drawRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color);
        void    fillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color);
        void    drawFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color);
        void    drawFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color);
        void    drawRGBBitmap(int16_t x, int16_t y, const uint16_t bitmap[], int16_t w, int16_t h, int16_t dir);
        void    setRotation(uint8_t r);
        void    invertDisplay(bool i);
        void    scrollTo(uint16_t y);

        // Transaction API not used by GFX
        void    setAddrWindow(uint16_t x, uint16_t y, uint16_t w, uint16_t h);

        uint8_t readcommand8(uint8_t reg, uint8_t index=0);

 private:
  uint8_t *buffer;
};

#endif // _ILI9163_buf_H_
