/*!
* ILI9341_game.cpp
*   ILI9341 TFT Displays with frame buffer
*
* Caution. This library use 153,600byte RAM for frame buffer
*
* Yuji Tanaka, Japan
*
*/

#include "ILI9341_game.h"
#ifndef ARDUINO_STM32_FEATHER
  #include "pins_arduino.h"
  #ifndef RASPI
    #include "wiring_private.h"
  #endif
#endif
#include <limits.h>

#if defined (ARDUINO_ARCH_ARC32) || defined (ARDUINO_MAXIM)
  #define SPI_DEFAULT_FREQ  16000000
#elif defined (__AVR__) || defined(TEENSYDUINO)
  #define SPI_DEFAULT_FREQ  8000000
#elif defined(ESP8266) || defined(ESP32)
  #define SPI_DEFAULT_FREQ  40000000
#elif defined(RASPI)
  #define SPI_DEFAULT_FREQ  80000000
#elif defined(ARDUINO_ARCH_STM32F1)
  #define SPI_DEFAULT_FREQ  36000000
#else
  #define SPI_DEFAULT_FREQ  24000000  ///< Default SPI data clock frequency
#endif

#define MADCTL_MY  0x80  ///< Bottom to top
#define MADCTL_MX  0x40  ///< Right to left
#define MADCTL_MV  0x20  ///< Reverse Mode
#define MADCTL_ML  0x10  ///< LCD refresh Bottom to top
#define MADCTL_RGB 0x00  ///< Red-Green-Blue pixel order
#define MADCTL_BGR 0x08  ///< Blue-Green-Red pixel order
#define MADCTL_MH  0x04  ///< LCD refresh right to left

#ifndef _swap_int16_t
#define _swap_int16_t(a, b) { int16_t t = a; a = b; b = t; }
#endif

/**************************************************************************/
/*!
    @brief  Instantiate Adafruit ILI9341 driver with software SPI
    @param    cs    Chip select pin #
    @param    dc    Data/Command pin #
    @param    mosi  SPI MOSI pin #
    @param    sclk  SPI Clock pin #
    @param    rst   Reset pin # (optional, pass -1 if unused)
    @param    miso  SPI MISO pin # (optional, pass -1 if unused)
*/
/**************************************************************************/
ILI9341_game::ILI9341_game(int8_t cs, int8_t dc, int8_t mosi,
        int8_t sclk, int8_t rst, int8_t miso) : Adafruit_SPITFT(ILI9341_TFTWIDTH, ILI9341_TFTHEIGHT, cs, dc, mosi, sclk, rst, miso) {
}

/**************************************************************************/
/*!
    @brief  Instantiate Adafruit ILI9341 driver with hardware SPI using the
            default SPI peripheral.
    @param  cs   Chip select pin # (OK to pass -1 if CS tied to GND).
    @param  dc   Data/Command pin # (required).
    @param  rst  Reset pin # (optional, pass -1 if unused).
*/
/**************************************************************************/
ILI9341_game::ILI9341_game(int8_t cs, int8_t dc, int8_t rst) :
  Adafruit_SPITFT(ILI9341_TFTWIDTH, ILI9341_TFTHEIGHT, cs, dc, rst) {
}

#if !defined(ESP8266)
/**************************************************************************/
/*!
    @brief  Instantiate Adafruit ILI9341 driver with hardware SPI using
            a specific SPI peripheral (not necessarily default).
    @param  spiClass  Pointer to SPI peripheral (e.g. &SPI or &SPI1).
    @param  dc        Data/Command pin # (required).
    @param  cs        Chip select pin # (optional, pass -1 if unused and
                      CS is tied to GND).
    @param  rst       Reset pin # (optional, pass -1 if unused).
*/
/**************************************************************************/
ILI9341_game::ILI9341_game(
  SPIClass *spiClass, int8_t dc, int8_t cs, int8_t rst) :
  Adafruit_SPITFT(ILI9341_TFTWIDTH, ILI9341_TFTHEIGHT, spiClass, cs, dc, rst) {
}
#endif // end !ESP8266

static const uint8_t PROGMEM initcmd[] = {
  0xEF, 3, 0x03, 0x80, 0x02,
  0xCF, 3, 0x00, 0xC1, 0x30,
  0xED, 4, 0x64, 0x03, 0x12, 0x81,
  0xE8, 3, 0x85, 0x00, 0x78,
  0xCB, 5, 0x39, 0x2C, 0x00, 0x34, 0x02,
  0xF7, 1, 0x20,
  0xEA, 2, 0x00, 0x00,
  ILI9341_PWCTR1  , 1, 0x23,             // Power control VRH[5:0]
  ILI9341_PWCTR2  , 1, 0x10,             // Power control SAP[2:0];BT[3:0]
  ILI9341_VMCTR1  , 2, 0x3e, 0x28,       // VCM control
//  ILI9341_VMCTR2  , 1, 0x86,             // VCM control2
  ILI9341_VMCTR2  , 1, 0x80,             // VCM control2
  ILI9341_MADCTL  , 1, 0x48,             // Memory Access Control
  ILI9341_VSCRSADD, 1, 0x00,             // Vertical scroll zero
  ILI9341_PIXFMT  , 1, 0x55,
  ILI9341_FRMCTR1 , 2, 0x00, 0x18,
  ILI9341_DFUNCTR , 3, 0x08, 0x82, 0x27, // Display Function Control
  0xF2, 1, 0x00,                         // 3Gamma Function Disable
  ILI9341_GAMMASET , 1, 0x01,             // Gamma curve selected
  ILI9341_GMCTRP1 , 15, 0x0F, 0x31, 0x2B, 0x0C, 0x0E, 0x08, // Set Gamma
    0x4E, 0xF1, 0x37, 0x07, 0x10, 0x03, 0x0E, 0x09, 0x00,
  ILI9341_GMCTRN1 , 15, 0x00, 0x0E, 0x14, 0x03, 0x11, 0x07, // Set Gamma
    0x31, 0xC1, 0x48, 0x08, 0x0F, 0x0C, 0x31, 0x36, 0x0F,
  ILI9341_SLPOUT  , 0x80,                // Exit Sleep
  ILI9341_DISPON  , 0x80,                // Display on
  0x00                                   // End of list
};

/**************************************************************************/
/*!
    @brief   Initialize ILI9341 chip
    Connects to the ILI9341 over SPI and sends initialization procedure commands
    @param    freq  Desired SPI clock frequency
*/
/**************************************************************************/
void ILI9341_game::begin(uint32_t freq) {

    if(!freq) freq = SPI_DEFAULT_FREQ;
    initSPI(freq);

    startWrite();

    if(_rst < 0) {                     // If no hardware reset pin...
        writeCommand(ILI9341_SWRESET); // Engage software reset
        delay(150);
    }

    uint8_t        cmd, x, numArgs;
    const uint8_t *addr = initcmd;
    while((cmd = pgm_read_byte(addr++)) > 0) {
        writeCommand(cmd);
        x       = pgm_read_byte(addr++);
        numArgs = x & 0x7F;
        while(numArgs--) spiWrite(pgm_read_byte(addr++));
        if(x & 0x80) delay(150);
    }

    endWrite();

    _width  = ILI9341_TFTWIDTH;
    _height = ILI9341_TFTHEIGHT;

    buffer[0] = (uint8_t *)malloc(_width * _height * 2 / 4);
    buffer[1] = (uint8_t *)malloc(_width * _height * 2 / 4);
    buffer[2] = (uint8_t *)malloc(_width * _height * 2 / 4);
    buffer[3] = (uint8_t *)malloc(_width * _height * 2 / 4);
    linebuf = (uint8_t *)malloc(_width * 2);
}


/**************************************************************************/
/*!
    @brief   Set origin of (0,0) and orientation of TFT display
    @param   m  The index for rotation, from 0-3 inclusive
*/
/**************************************************************************/
void ILI9341_game::setRotation(uint8_t m) {
    rotation = m % 4; // can't be higher than 3
    switch (rotation) {
        case 0:
            m = (MADCTL_MX | MADCTL_BGR);
            _width  = ILI9341_TFTWIDTH;
            _height = ILI9341_TFTHEIGHT;
            break;
        case 1:
            m = (MADCTL_MV | MADCTL_BGR);
            _width  = ILI9341_TFTHEIGHT;
            _height = ILI9341_TFTWIDTH;
            break;
        case 2:
            m = (MADCTL_MY | MADCTL_BGR);
            _width  = ILI9341_TFTWIDTH;
            _height = ILI9341_TFTHEIGHT;
            break;
        case 3:
            m = (MADCTL_MX | MADCTL_MY | MADCTL_MV | MADCTL_BGR);
            _width  = ILI9341_TFTHEIGHT;
            _height = ILI9341_TFTWIDTH;
            break;
    }

    startWrite();
    writeCommand(ILI9341_MADCTL);
    spiWrite(m);
    endWrite();
}

/**************************************************************************/
/*!
    @brief   Enable/Disable display color inversion
    @param   invert True to invert, False to have normal color
*/
/**************************************************************************/
void ILI9341_game::invertDisplay(bool invert) {
    startWrite();
    writeCommand(invert ? ILI9341_INVON : ILI9341_INVOFF);
    endWrite();
}

/**************************************************************************/
/*!
    @brief   Scroll display memory
    @param   y How many pixels to scroll display by
*/
/**************************************************************************/
void ILI9341_game::scrollTo(uint16_t y) {
    startWrite();
    writeCommand(ILI9341_VSCRSADD);
    SPI_WRITE16(y);
    endWrite();
}

/**************************************************************************/
/*!
    @brief   Set the "address window" - the rectangle we will write to RAM with the next chunk of SPI data writes. The ILI9341 will automatically wrap the data as each row is filled
    @param   x1  TFT memory 'x' origin
    @param   y1  TFT memory 'y' origin
    @param   w   Width of rectangle
    @param   h   Height of rectangle
*/
/**************************************************************************/
void ILI9341_game::setAddrWindow(
  uint16_t x1, uint16_t y1, uint16_t w, uint16_t h) {
    uint16_t x2 = (x1 + w - 1),
             y2 = (y1 + h - 1);
    writeCommand(ILI9341_CASET); // Column address set
    SPI_WRITE16(x1);
    SPI_WRITE16(x2);
    writeCommand(ILI9341_PASET); // Row address set
    SPI_WRITE16(y1);
    SPI_WRITE16(y2);
    writeCommand(ILI9341_RAMWR); // Write to RAM
}

/**************************************************************************/
/*!
   @brief  Read 8 bits of data from ILI9341 configuration memory. NOT from RAM!
           This is highly undocumented/supported, it's really a hack but kinda works?
    @param    command  The command register to read data from
    @param    index  The byte index into the command to read from
    @return   Unsigned 8-bit data read from ILI9341 register
*/
/**************************************************************************/
uint8_t ILI9341_game::readcommand8(uint8_t command, uint8_t index) {
    startWrite();
    writeCommand(0xD9);  // woo sekret command?
    spiWrite(0x10 + index);
    writeCommand(command);
    uint8_t r = spiRead();
    endWrite();
    return r;
}
void ILI9341_game::display() {
    startWrite();
    setAddrWindow(0, 0, _width, _height);
    writePixels((uint16_t *)buffer[0], _width * _height / 4);
    writePixels((uint16_t *)buffer[1], _width * _height / 4);
    writePixels((uint16_t *)buffer[2], _width * _height / 4);
    writePixels((uint16_t *)buffer[3], _width * _height / 4);
    endWrite();
}

void ILI9341_game::display2x() {
    uint16_t i, j;
    uint32_t ww, hh;

    ww=_width / 2;
    hh=_height / 4;
    startWrite();
    setAddrWindow(0, 0, _width, _height);

    for(i=0; i<hh; i++){
        for(j=0; j<ww; j++){
            linebuf[j*4  ]=buffer[0][i*_width*2+j*2];
            linebuf[j*4+1]=buffer[0][i*_width*2+j*2+1];
            linebuf[j*4+2]=buffer[0][i*_width*2+j*2];
            linebuf[j*4+3]=buffer[0][i*_width*2+j*2+1];
        }
        writePixels((uint16_t *)linebuf, _width);
        writePixels((uint16_t *)linebuf, _width);
    }

    for(i=0; i<hh; i++){
        for(j=0; j<ww; j++){
            linebuf[j*4  ]=buffer[1][i*_width*2+j*2];
            linebuf[j*4+1]=buffer[1][i*_width*2+j*2+1];
            linebuf[j*4+2]=buffer[1][i*_width*2+j*2];
            linebuf[j*4+3]=buffer[1][i*_width*2+j*2+1];
        }
        writePixels((uint16_t *)linebuf, _width);
        writePixels((uint16_t *)linebuf, _width);
    }
    endWrite();
}

void ILI9341_game::clearDisplay() {
    memset(buffer[0], 0, _width * _height * 2 / 4);
    memset(buffer[1], 0, _width * _height * 2 / 4);
    memset(buffer[2], 0, _width * _height * 2 / 4);
    memset(buffer[3], 0, _width * _height * 2 / 4);
}

void ILI9341_game::writePixel(int16_t x, int16_t y, uint16_t color) {
    uint16_t i;
    uint16_t m, n;
    uint8_t *tmppt;

    if (x<0 || x>=_width || y<0 || y>=_height)
        return;

    m=y * 4 / _height;
    n=y - m * _height / 4;

    i=((n * _width) + x) * 2;
    tmppt=&buffer[m][i];
    *(tmppt++)=color & 0xFF;
    *tmppt=color >> 8;
}

void ILI9341_game::writeFillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color) {
    uint16_t i;
    uint16_t m, n;
    int16_t  x0, y0;
    int16_t  x1, y1;
    int16_t  px, py;
    uint8_t *tmppt;

    if ((x+w)<0 || x>=_width || (y+h)<0 || y>=_height)
        return;

    x0=x;
    x1=x+w;
    y0=y;
    y1=y+h;

    if (x<0)
      x0=0;
    if ((x+w)>_width)
      x1=_width;

    if (y<0)
      y0=0;
    if ((y+h)>_height)
      y1=_height;

    for(py=y0; py<y1; py++){
      m=py * 4 / _height;
      n=py - m * _height / 4;
      for(px=x0; px<x1; px++){
        i=((n * _width) + px) * 2;
        tmppt=&buffer[m][i];
        *(tmppt++)=color & 0xFF;
        *tmppt=color >> 8;
      }
    }
}

void ILI9341_game::writeFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color) {
    uint16_t i;
    uint16_t m, n;
    int16_t  y0, y1;
    int16_t  px, py;
    uint8_t *tmppt;

    if (x<0 || x>=_width || (y+h)<0 || y>=_height)
        return;

    y0=y;
    y1=y+h;

    if (y<0)
      y0=0;
    if ((y+h)>_height)
      y1=_height;

    px=x;
    for(py=y0; py<y1; py++){
        m=py * 4 / _height;
        n=py - m * _height / 4;

        i=((n * _width) + px) * 2;
        tmppt=&buffer[m][i];
        *(tmppt++)=color & 0xFF;
        *tmppt=color >> 8;
    }
}

void ILI9341_game::writeFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color) {
    uint16_t i;
    uint16_t m, n;
    int16_t  x0, x1;
    int16_t  px, py;
    uint8_t *tmppt;

    if ((x+w)<0 || x>=_width || y<0 || y>=_height)
        return;

    x0=x;
    x1=x+w;

    if (x<0)
      x0=0;
    if ((x+w)>_width)
      x1=_width;

    py=y;
    m=py * 4 / _height;
    n=py - m * _height / 4;
    for(px=x0; px<x1; px++){
        i=((n * _width) + px) * 2;
        tmppt=&buffer[m][i];
        *(tmppt++)=color & 0xFF;
        *tmppt=color >> 8;
    }
}

void ILI9341_game::drawPixel(int16_t x, int16_t y, uint16_t color) {
    writePixel(x, y, color);
}

void ILI9341_game::drawRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color) {
    writeFastVLine(x, y, h, color);
    writeFastVLine(x + w, y, h, color);
    writeFastHLine(x, y, w, color);
    writeFastHLine(x, y + h, w, color);
}

void ILI9341_game::fillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color) {
    writeFillRect(x, y, w, h, color);
}

void ILI9341_game::drawFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color) {
    writeFastVLine(x, y, h, color);
}

void ILI9341_game::drawFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color) {
    writeFastHLine(x, y, w, color);
}

void ILI9341_game::drawRGBBitmap(int16_t x, int16_t y, const uint16_t bitmap[], int16_t w, int16_t h, int16_t dir) {
	int inside;
	int bufn, bufy, bufs;
	const uint16_t *tmppt;
	uint8_t *buf;
	uint16_t val;

	if ((x+w)<0 || x>=_width || (y+h)<0 || y>=_height)
		return;

	if (x<0 || (x+w)>=_width || y<0 || (y+h)>=_height)
		inside=0;
	else
		inside=1;

	tmppt = bitmap;
	bufs = _height / 4;
	if (dir==0 || dir==14){
		if (inside==1){
			for(int16_t j=0; j<h; j++, y++) {
				bufn=y / bufs;
				bufy=y % bufs;
				buf=&buffer[bufn][(bufy*_width + x) * 2];
				for(int16_t i=0; i<w; i++ ) {
					val=*(tmppt++);
					*(buf++)=val & 0xFF;
					*(buf++)=val >> 8;
				}
			}
		}else{
			for(int16_t j=0; j<h; j++, y++) {
				for(int16_t i=0; i<w; i++ ) {
					writePixel(x+i, y, *(tmppt++));
				}
			}
		}
	}
	if (dir==1 || dir==15){
		if (inside==1){
			for(int16_t j=0; j<w; j++, y++) {
				bufn=y / bufs;
				bufy=y % bufs;
				buf=&buffer[bufn][(bufy*_width + x) * 2];
				for(int16_t i=0; i<h; i++ ) {
					val=tmppt[(h-1-i)*w+j];
					*(buf++)=val & 0xFF;
					*(buf++)=val >> 8;
				}
			}
		}else{
			for(int16_t j=0; j<w; j++, y++) {
				for(int16_t i=0; i<h; i++ ) {
					writePixel(x+i, y, tmppt[(h-1-i)*w+j]);
				}
			}
		}
	}
	if (dir==2 || dir==12){
		if (inside==1){
			for(int16_t j=0; j<h; j++) {
				bufn=(y+h-1-j) / bufs;
				bufy=(y+h-1-j) % bufs;
				buf=&buffer[bufn][(bufy*_width + x+w-1) * 2];
				for(int16_t i=0; i<w; i++ ) {
					val=*(tmppt++);
					*(buf++)=val & 0xFF;
					*(buf++)=val >> 8;
					buf-=4;
				}
			}
		}else{
			for(int16_t j=0; j<h; j++) {
				for(int16_t i=0; i<w; i++ ) {
					writePixel(x+w-1-i, y+h-1-j, *(tmppt++));
				}
			}
		}
	}
	if (dir==3 || dir==13){
		if (inside==1){
			for(int16_t j=0; j<w; j++, y++) {
				bufn=y / bufs;
				bufy=y % bufs;
				buf=&buffer[bufn][(bufy*_width + x) * 2];
				for(int16_t i=0; i<h; i++ ) {
					val=tmppt[i*w+h-1-j];
					*(buf++)=val & 0xFF;
					*(buf++)=val >> 8;
				}
			}
		}else{
			for(int16_t j=0; j<w; j++, y++) {
				for(int16_t i=0; i<h; i++ ) {
					writePixel(x+i, y, tmppt[i*w+h-1-j]);
				}
			}
		}
	}
	if (dir==4 || dir==10){
		if (inside==1){
			for(int16_t j=0; j<h; j++, y++) {
				bufn=y / bufs;
				bufy=y % bufs;
				buf=&buffer[bufn][(bufy*_width + x+w-1) * 2];
				for(int16_t i=0; i<w; i++ ) {
					val=*(tmppt++);
					*(buf++)=val & 0xFF;
					*(buf++)=val >> 8;
					buf-=4;
				}
			}
		}else{
			for(int16_t j=0; j<h; j++, y++) {
				for(int16_t i=0; i<w; i++ ) {
					writePixel(x+w-1-i, y, *(tmppt++));
				}
			}
		}
	}
	if (dir==5 || dir==11){
		if (inside==1){
			for(int16_t j=0; j<w; j++, y++) {
				bufn=y / bufs;
				bufy=y % bufs;
				buf=&buffer[bufn][(bufy*_width + x+w-1) * 2];
				for(int16_t i=0; i<h; i++ ) {
					val=tmppt[i*w+w-1-j];
					*(buf++)=val & 0xFF;
					*(buf++)=val >> 8;
					buf-=4;
				}
			}
		}else{
			for(int16_t j=0; j<w; j++, y++) {
				for(int16_t i=0; i<h; i++ ) {
					writePixel(x+w-1-i, y, tmppt[i*w+w-1-j]);
				}
			}
		}
	}
	if (dir==6 || dir==8){
		if (inside==1){
			for(int16_t j=0; j<h; j++) {
				bufn=(y+h-1-j) / bufs;
				bufy=(y+h-1-j) % bufs;
				buf=&buffer[bufn][(bufy*_width + x) * 2];
				for(int16_t i=0; i<w; i++ ) {
					val=*(tmppt++);
					*(buf++)=val & 0xFF;
					*(buf++)=val >> 8;
				}
			}
		}else{
			for(int16_t j=0; j<h; j++) {
				for(int16_t i=0; i<w; i++ ) {
					writePixel(x+i, y+h-1-j, *(tmppt++));
				}
			}
		}
	}
	if (dir==7 || dir==9){
		if (inside==1){
			for(int16_t j=0; j<w; j++, y++) {
				bufn=y / bufs;
				bufy=y % bufs;
				buf=&buffer[bufn][(bufy*_width + x+w-1) * 2];
				for(int16_t i=0; i<h; i++ ) {
					val=tmppt[(h-1-i)*w+j];
					*(buf++)=val & 0xFF;
					*(buf++)=val >> 8;
					buf-=4;
				}
			}
		}else{
			for(int16_t j=0; j<w; j++, y++) {
				for(int16_t i=0; i<h; i++ ) {
					writePixel(x+w-1-i, y, tmppt[(h-1-i)*w+j]);
				}
			}
		}
	}
}

void ILI9341_game::drawRGBBitmap2x(int16_t x, int16_t y, const uint16_t bitmap[], int16_t w, int16_t h, int16_t dir) {
	uint16_t val;
	const uint16_t *tmppt;

	tmppt = bitmap;
	if (dir==0 || dir==14){
		for(int16_t j=0; j<h; j++, y+=2) {
			for(int16_t i=0; i<w; i++ ) {
				val=*(tmppt++);
				writePixel(x+i*2, y, val);
				writePixel(x+i*2+1, y, val);
				writePixel(x+i*2, y+1, val);
				writePixel(x+i*2+1, y+1, val);
			}
		}
	}
	if (dir==1 || dir==15){
		for(int16_t j=0; j<w; j++, y+=2) {
			for(int16_t i=0; i<h; i++ ) {
				val=tmppt[(h-1-i)*w+j];
				writePixel(x+i*2, y, val);
				writePixel(x+i*2+1, y, val);
				writePixel(x+i*2, y+1, val);
				writePixel(x+i*2+1, y+1, val);
			}
		}
	}
	if (dir==2 || dir==12){
		for(int16_t j=0; j<h; j++, y+=2) {
			for(int16_t i=0; i<w; i++ ) {
				val=tmppt[(h-1-j)*w+w-1-i];
				writePixel(x+i*2, y, val);
				writePixel(x+i*2+1, y, val);
				writePixel(x+i*2, y+1, val);
				writePixel(x+i*2+1, y+1, val);
			}
		}
	}
	if (dir==3 || dir==13){
		for(int16_t j=0; j<w; j++, y+=2) {
			for(int16_t i=0; i<h; i++ ) {
				val=tmppt[i*w+w-1-j];
				writePixel(x+i*2, y, val);
				writePixel(x+i*2+1, y, val);
				writePixel(x+i*2, y+1, val);
				writePixel(x+i*2+1, y+1, val);
			}
		}
	}
	if (dir==4 || dir==10){
		for(int16_t j=0; j<h; j++, y+=2) {
			for(int16_t i=0; i<w; i++ ) {
				val=tmppt[j*w+w-1-i];
				writePixel(x+i*2, y, val);
				writePixel(x+i*2+1, y, val);
				writePixel(x+i*2, y+1, val);
				writePixel(x+i*2+1, y+1, val);
			}
		}
	}
	if (dir==5 || dir==11){
		for(int16_t j=0; j<w; j++, y+=2) {
			for(int16_t i=0; i<h; i++ ) {
				val=tmppt[(h-1-i)*w+w-1-j];
				writePixel(x+i*2, y, val);
				writePixel(x+i*2+1, y, val);
				writePixel(x+i*2, y+1, val);
				writePixel(x+i*2+1, y+1, val);
			}
		}
	}
	if (dir==6 || dir==8){
		for(int16_t j=0; j<h; j++, y+=2) {
			for(int16_t i=0; i<w; i++ ) {
				val=tmppt[(h-1-j)*w+i];
				writePixel(x+i*2, y, val);
				writePixel(x+i*2+1, y, val);
				writePixel(x+i*2, y+1, val);
				writePixel(x+i*2+1, y+1, val);
			}
		}
	}
	if (dir==7 || dir==9){
		for(int16_t j=0; j<w; j++, y+=2) {
			for(int16_t i=0; i<h; i++ ) {
				val=tmppt[i*w+j];
				writePixel(x+i*2, y, val);
				writePixel(x+i*2+1, y, val);
				writePixel(x+i*2, y+1, val);
				writePixel(x+i*2+1, y+1, val);
			}
		}
	}
}

void ILI9341_game::drawRGBSprite(int16_t x, int16_t y, const uint16_t bitmap[], int16_t w, int16_t h, int16_t dir) {
	int inside;
	int bufn, bufy, bufs;
	const uint16_t *tmppt;
	uint8_t *buf;
	uint16_t val;

	if ((x+w)<0 || x>=_width || (y+h)<0 || y>=_height)
		return;

	if (x<0 || (x+w)>=_width || y<0 || (y+h)>=_height)
		inside=0;
	else
		inside=1;

	tmppt = bitmap;
	bufs = _height / 4;
	if (dir==0 || dir==14){
		if (inside==1){
			for(int16_t j=0; j<h; j++, y++) {
				bufn=y / bufs;
				bufy=y % bufs;
				buf=&buffer[bufn][(bufy*_width + x) * 2];
				for(int16_t i=0; i<w; i++ ) {
					val=*(tmppt++);
					if (val!=ILI9341_THROUGH){
						*(buf++)=val & 0xFF;
						*(buf++)=val >> 8;
					}else{
						buf+=2;
					}
				}
			}
		}else{
			for(int16_t j=0; j<h; j++, y++) {
				for(int16_t i=0; i<w; i++ ) {
					val=*(tmppt++);
					if (val!=ILI9341_THROUGH){
						writePixel(x+i, y, val);
					}
				}
			}
		}
	}
	if (dir==1 || dir==15){
		if (inside==1){
			for(int16_t j=0; j<w; j++, y++) {
				bufn=y / bufs;
				bufy=y % bufs;
				buf=&buffer[bufn][(bufy*_width + x) * 2];
				for(int16_t i=0; i<h; i++ ) {
					val=tmppt[(h-1-i)*w+j];
					if (val!=ILI9341_THROUGH){
						*(buf++)=val & 0xFF;
						*(buf++)=val >> 8;
					}else{
						buf+=2;
					}
				}
			}
		}else{
			for(int16_t j=0; j<w; j++, y++) {
				for(int16_t i=0; i<h; i++ ) {
					val=tmppt[(h-1-i)*w+j];
					if (val!=ILI9341_THROUGH){
						writePixel(x+i, y, val);
					}
				}
			}
		}
	}
	if (dir==2 || dir==12){
		if (inside==1){
			for(int16_t j=0; j<h; j++) {
				bufn=(y+h-1-j) / bufs;
				bufy=(y+h-1-j) % bufs;
				buf=&buffer[bufn][(bufy*_width + x+w-1) * 2];
				for(int16_t i=0; i<w; i++ ) {
					val=*(tmppt++);
					if (val!=ILI9341_THROUGH){
						*(buf++)=val & 0xFF;
						*(buf++)=val >> 8;
						buf-=4;
					}else{
						buf-=2;
					}
				}
			}
		}else{
			for(int16_t j=0; j<h; j++) {
				for(int16_t i=0; i<w; i++ ) {
					val=*(tmppt++);
					if (val!=ILI9341_THROUGH){
						writePixel(x+w-1-i, y+h-1-j, val);
					}
				}
			}
		}
	}
	if (dir==3 || dir==13){
		if (inside==1){
			for(int16_t j=0; j<w; j++, y++) {
				bufn=y / bufs;
				bufy=y % bufs;
				buf=&buffer[bufn][(bufy*_width + x) * 2];
				for(int16_t i=0; i<h; i++ ) {
					val=tmppt[i*w+h-1-j];
					if (val!=ILI9341_THROUGH){
						*(buf++)=val & 0xFF;
						*(buf++)=val >> 8;
					}else{
						buf+=2;
					}
				}
			}
		}else{
			for(int16_t j=0; j<w; j++, y++) {
				for(int16_t i=0; i<h; i++ ) {
					val=tmppt[i*w+h-1-j];
					if (val!=ILI9341_THROUGH){
						writePixel(x+i, y, val);
					}
				}
			}
		}
	}
	if (dir==4 || dir==10){
		if (inside==1){
			for(int16_t j=0; j<h; j++, y++) {
				bufn=y / bufs;
				bufy=y % bufs;
				buf=&buffer[bufn][(bufy*_width + x+w-1) * 2];
				for(int16_t i=0; i<w; i++ ) {
					val=*(tmppt++);
					if (val!=ILI9341_THROUGH){
						*(buf++)=val & 0xFF;
						*(buf++)=val >> 8;
						buf-=4;
					}else{
						buf-=2;
					}
				}
			}
		}else{
			for(int16_t j=0; j<h; j++, y++) {
				for(int16_t i=0; i<w; i++ ) {
					val=*(tmppt++);
					if (val!=ILI9341_THROUGH){
						writePixel(x+w-1-i, y, val);
					}
				}
			}
		}
	}
	if (dir==5 || dir==11){
		if (inside==1){
			for(int16_t j=0; j<w; j++, y++) {
				bufn=y / bufs;
				bufy=y % bufs;
				buf=&buffer[bufn][(bufy*_width + x+w-1) * 2];
				for(int16_t i=0; i<h; i++ ) {
					val=tmppt[i*w+w-1-j];
					if (val!=ILI9341_THROUGH){
						*(buf++)=val & 0xFF;
						*(buf++)=val >> 8;
						buf-=4;
					}else{
						buf-=2;
					}
				}
			}
		}else{
			for(int16_t j=0; j<w; j++, y++) {
				for(int16_t i=0; i<h; i++ ) {
					val=tmppt[i*w+w-1-j];
					if (val!=ILI9341_THROUGH){
						writePixel(x+w-1-i, y, val);
					}
				}
			}
		}
	}
	if (dir==6 || dir==8){
		if (inside==1){
			for(int16_t j=0; j<h; j++) {
				bufn=(y+h-1-j) / bufs;
				bufy=(y+h-1-j) % bufs;
				buf=&buffer[bufn][(bufy*_width + x) * 2];
				for(int16_t i=0; i<w; i++ ) {
					val=*(tmppt++);
					if (val!=ILI9341_THROUGH){
						*(buf++)=val & 0xFF;
						*(buf++)=val >> 8;
					}else{
						buf+=2;
					}
				}
			}
		}else{
			for(int16_t j=0; j<h; j++) {
				for(int16_t i=0; i<w; i++ ) {
					val=*(tmppt++);
					if (val!=ILI9341_THROUGH){
						writePixel(x+i, y+h-1-j, *(tmppt++));
					}
				}
			}
		}
	}
	if (dir==7 || dir==9){
		if (inside==1){
			for(int16_t j=0; j<w; j++, y++) {
				bufn=y / bufs;
				bufy=y % bufs;
				buf=&buffer[bufn][(bufy*_width + x+w-1) * 2];
				for(int16_t i=0; i<h; i++ ) {
					val=tmppt[(h-1-i)*w+j];
					if (val!=ILI9341_THROUGH){
						*(buf++)=val & 0xFF;
						*(buf++)=val >> 8;
						buf-=4;
					}else{
						buf-=2;
					}
				}
			}
		}else{
			for(int16_t j=0; j<w; j++, y++) {
				for(int16_t i=0; i<h; i++ ) {
					val=tmppt[(h-1-i)*w+j];
					if (val!=ILI9341_THROUGH){
						writePixel(x+w-1-i, y, val);
					}
				}
			}
		}
	}
}

void ILI9341_game::drawRGBSprite2x(int16_t x, int16_t y, const uint16_t bitmap[], int16_t w, int16_t h, int16_t dir) {
	uint16_t val;
	const uint16_t *tmppt;

	tmppt = bitmap;
	if (dir==0 || dir==14){
		for(int16_t j=0; j<h; j++, y+=2) {
			for(int16_t i=0; i<w; i++ ) {
				val=*(tmppt++);
				if (val!=ILI9341_THROUGH){
					writePixel(x+i*2, y, val);
					writePixel(x+i*2+1, y, val);
					writePixel(x+i*2, y+1, val);
					writePixel(x+i*2+1, y+1, val);
				}
			}
		}
	}
	if (dir==1 || dir==15){
		for(int16_t j=0; j<w; j++, y+=2) {
			for(int16_t i=0; i<h; i++ ) {
				val=tmppt[(h-1-i)*w+j];
				if (val!=ILI9341_THROUGH){
					writePixel(x+i*2, y, val);
					writePixel(x+i*2+1, y, val);
					writePixel(x+i*2, y+1, val);
					writePixel(x+i*2+1, y+1, val);
				}
			}
		}
	}
	if (dir==2 || dir==12){
		for(int16_t j=0; j<h; j++, y+=2) {
			for(int16_t i=0; i<w; i++ ) {
				val=tmppt[(h-1-j)*w+w-1-i];
				if (val!=ILI9341_THROUGH){
					writePixel(x+i*2, y, val);
					writePixel(x+i*2+1, y, val);
					writePixel(x+i*2, y+1, val);
					writePixel(x+i*2+1, y+1, val);
				}
			}
		}
	}
	if (dir==3 || dir==13){
		for(int16_t j=0; j<w; j++, y+=2) {
			for(int16_t i=0; i<h; i++ ) {
				val=tmppt[i*w+w-1-j];
				if (val!=ILI9341_THROUGH){
					writePixel(x+i*2, y, val);
					writePixel(x+i*2+1, y, val);
					writePixel(x+i*2, y+1, val);
					writePixel(x+i*2+1, y+1, val);
				}
			}
		}
	}
	if (dir==4 || dir==10){
		for(int16_t j=0; j<h; j++, y+=2) {
			for(int16_t i=0; i<w; i++ ) {
				val=tmppt[j*w+w-1-i];
				if (val!=ILI9341_THROUGH){
					writePixel(x+i*2, y, val);
					writePixel(x+i*2+1, y, val);
					writePixel(x+i*2, y+1, val);
					writePixel(x+i*2+1, y+1, val);
				}
			}
		}
	}
	if (dir==5 || dir==11){
		for(int16_t j=0; j<w; j++, y+=2) {
			for(int16_t i=0; i<h; i++ ) {
				val=tmppt[(h-1-i)*w+w-1-j];
				if (val!=ILI9341_THROUGH){
					writePixel(x+i*2, y, val);
					writePixel(x+i*2+1, y, val);
					writePixel(x+i*2, y+1, val);
					writePixel(x+i*2+1, y+1, val);
				}
			}
		}
	}
	if (dir==6 || dir==8){
		for(int16_t j=0; j<h; j++, y+=2) {
			for(int16_t i=0; i<w; i++ ) {
				val=tmppt[(h-1-j)*w+i];
				if (val!=ILI9341_THROUGH){
					writePixel(x+i*2, y, val);
					writePixel(x+i*2+1, y, val);
					writePixel(x+i*2, y+1, val);
					writePixel(x+i*2+1, y+1, val);
				}
			}
		}
	}
	if (dir==7 || dir==9){
		for(int16_t j=0; j<w; j++, y+=2) {
			for(int16_t i=0; i<h; i++ ) {
				val=tmppt[i*w+j];
				if (val!=ILI9341_THROUGH){
					writePixel(x+i*2, y, val);
					writePixel(x+i*2+1, y, val);
					writePixel(x+i*2, y+1, val);
					writePixel(x+i*2+1, y+1, val);
				}
			}
		}
	}
}
